<?php

namespace Drupal\Tests\tome_static\Kernel;

use Drupal\block_content\Entity\BlockContent;
use Drupal\block_content\Entity\BlockContentType;
use Drupal\Core\Datetime\Entity\DateFormat;
use Drupal\Core\Site\Settings;
use Drupal\node\Entity\Node;
use Drupal\Tests\tome_base\Kernel\TestBase;

/**
 * Tests that static site generation works.
 *
 * @coversDefaultClass \Drupal\tome_static\StaticGenerator
 * @group tome_static
 */
class StaticGeneratorTest extends TestBase {

  /**
   * {@inheritdoc}
   */
  public static $modules = [
    'tome_static',
  ];

  /**
   * @covers \Drupal\tome_static\StaticGenerator::getPaths
   * @covers \Drupal\tome_static\EventSubscriber\LanguagePathSubscriber::collectPaths
   * @covers \Drupal\tome_static\EventSubscriber\EntityPathSubscriber::collectPaths
   * @covers \Drupal\tome_static\EventSubscriber\EntityPathSubscriber::replacePathPlaceholder
   */
  public function testGetPaths() {
    /** @var \Drupal\tome_static\StaticGenerator $static */
    $static = \Drupal::service('tome_static.generator');
    $base_dir = Settings::get('tome_static_directory');

    DateFormat::create([
      'pattern' => 'D, m/d/Y - H:i',
      'id' => 'medium',
    ])->save();

    $this->assertUnsortedEquals([
      '/',
      '/french',
    ], $static->getPaths());

    $article = Node::create(['type' => 'article', 'title' => 'My article']);
    $article->addTranslation('fr', ['title' => 'My french article']);
    $article->save();

    $this->assertUnsortedEquals([
      '/',
      '_entity:node:en:1',
      '_entity:node:fr:1',
      '/french',
    ], $static->getPaths());

    $settings = Settings::getInstance() ? Settings::getAll() : [];
    $settings['tome_static_path_exclude'] = ['_entity:node:fr:1', '/french'];
    new Settings($settings);

    $this->assertUnsortedEquals([
      '/',
      '_entity:node:en:1',
    ], $static->getPaths());

    $static->requestPath('_entity:node:en:1');
    $this->assertFileExists("$base_dir/node/1/index.html");
    $static->requestPath('_entity:node:fr:1');
    $this->assertFileExists("$base_dir/french/node/1/index.html");
  }

  /**
   * @covers \Drupal\tome_static\EventSubscriber\LanguagePathSubscriber::collectPaths
   */
  public function testLanguagePathPrefix() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome_static\StaticGenerator $static */
    $static = \Drupal::service('tome_static.generator');

    $this->assertUnsortedEquals([
      '/',
      '/tome-test/page',
      '/tome-test/complex-page',
      '/tome-test/pager-page',
      '/french',
      '/french/tome-test/page',
      '/french/tome-test/complex-page',
      '/french/tome-test/pager-page',
    ], $static->getPaths());
  }

  /**
   * @covers \Drupal\tome_static\StaticGenerator::requestPath
   */
  public function testRequestPath() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome_static\StaticGenerator $static */
    $static = \Drupal::service('tome_static.generator');

    $static->requestPath('/tome-test/page');
    $this->assertContains('Hello, world', file_get_contents(Settings::get('tome_static_directory') . '/tome-test/page/index.html'));
  }

  /**
   * @covers \Drupal\tome_static\StaticGenerator::requestPath
   * @covers \Drupal\tome_static\StaticGenerator::getCssAssets
   * @covers \Drupal\tome_static\StaticGenerator::getHtmlAssets
   * @covers \Drupal\tome_static\StaticGenerator::exportPaths
   */
  public function testComplexHtml() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome_static\StaticGenerator $static */
    $static = \Drupal::service('tome_static.generator');

    $invoke_paths = $static->requestPath('/tome-test/complex-page');
    $module_path = drupal_get_path('module', 'tome_test');

    $this->assertContains("/$module_path/assets/srcset-test1.png", $invoke_paths);
    $this->assertContains("/$module_path/assets/srcset-test2.png", $invoke_paths);
    $this->assertContains("/$module_path/assets/srcset-test3.png", $invoke_paths);

    $static->exportPaths($invoke_paths);
    $base_dir = Settings::get('tome_static_directory');
    $this->assertContains('Hello, complex world', file_get_contents("$base_dir/tome-test/complex-page/index.html"));
    $this->assertFileExists("$base_dir/$module_path/css/complex.css");
    $this->assertFileExists("$base_dir/$module_path/js/complex.js");
    $this->assertFileExists("$base_dir/$module_path/assets/druplicon-1.png");
    $this->assertFileExists("$base_dir/$module_path/assets/file with space.png");
    $this->assertFileExists("$base_dir/core/misc/druplicon.png");
  }

  /**
   * @covers \Drupal\tome_static\StaticGenerator::requestPath
   * @covers \Drupal\tome_static\StaticGenerator::replacePagerUrl
   * @covers \Drupal\tome_static\StaticGenerator::handlePagers
   * @covers \Drupal\tome_static\StaticGenerator::exportPaths
   */
  public function testPagers() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome_static\StaticGenerator $static */
    $static = \Drupal::service('tome_static.generator');

    $invoke_paths = $static->requestPath('/tome-test/pager-page');
    $this->assertContains('/tome-test/pager-page?page=0', $invoke_paths);
    $this->assertContains('/tome-test/pager-page?page=1', $invoke_paths);
    $this->assertContains('/foobar?page=2', $invoke_paths);
    $base_dir = Settings::get('tome_static_directory');
    $contents = file_get_contents("$base_dir/tome-test/pager-page/index.html");
    $this->assertContains('href="/tome-test/pager-page"', $contents);
    $this->assertContains('href="/tome-test/pager-page/page/2"', $contents);
    $this->assertContains('href="/foobar/page/3"', $contents);
  }

  /**
   * @covers \Drupal\tome_static\EventSubscriber\EntityPathSubscriber::collectPaths
   */
  public function testExcludedEntityTypes() {
    $this->enableModules(['link', 'block_content']);
    $this->installEntitySchema('block_content');

    /** @var \Drupal\tome_static\StaticGenerator $static */
    $static = \Drupal::service('tome_static.generator');

    BlockContentType::create([
      'id' => 'basic',
      'label' => 'basic',
    ])->save();

    BlockContent::create([
      'info' => 'block',
      'type' => 'basic',
      'langcode' => 'en',
    ])->save();

    $this->assertNotContains('_entity:block_content:en:1', $static->getPaths());
  }

  /**
   * Asserts that two arrays are equal, ignoring sort order.
   *
   * @param array $array1
   *   The first array to compare.
   * @param array $array2
   *   The second array to compare.
   */
  protected function assertUnsortedEquals(array $array1, array $array2) {
    sort($array1);
    sort($array2);
    $this->assertEquals($array1, $array2);
  }

}
